"""Support sorting for Mongo ORM."""
from __future__ import annotations

from typing import TYPE_CHECKING

from muffin_rest.sorting import Sort, Sorting

if TYPE_CHECKING:
    from .types import TVCollection


class MongoSort(Sort):
    """Sorter for Peewee."""

    async def apply(self, collection, *, desc: bool = False, **_):
        """Sort the collection."""
        collection.sorting.append((self.field, -1 if desc else 1))
        return collection


class MongoSorting(Sorting):
    """Sort Peewee ORM Queries."""

    MUTATE_CLASS = MongoSort

    def sort_default(self, collection: TVCollection) -> TVCollection:
        """Sort collection by default."""
        res = collection.sort(
            [
                (sort.name, -1 if sort.meta["default"] == "desc" else 1)
                for sort in self.default
            ],
        )
        return res



"""Create openapi schema from the given API."""
import inspect
import re
from contextlib import suppress
from functools import partial
from http import HTTPStatus
from typing import Callable, Dict, List, Tuple, cast

from apispec import utils
from apispec.core import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from asgi_tools.response import CAST_RESPONSE
from asgi_tools.types import TJSON
from http_router.routes import DynamicRoute, Route
from muffin import Response

from . import LIMIT_PARAM, OFFSET_PARAM, openapi
from .options import RESTOptions

with suppress(ImportError):
    from apispec import yaml_utils

locals().setdefault("yaml_utils", None)


DEFAULT_METHODS = ("get",)
HTTP_METHODS = [
    "GET",
    "POST",
    "PUT",
    "PATH",
    "DELETE",
    "HEAD",
    "OPTIONS",
    "TRACE",
    "CONNECT",
]
RE_URL = re.compile(r"<(?:[^:<>]+:)?([^<>]+)>")
SKIP_PATH = {"/openapi.json", "/swagger", "/redoc"}


def render_openapi(api, request):
    """Prepare openapi specs."""
    # Setup Specs
    options = dict(api.openapi_options)
    options.setdefault(
        "servers",
        [{"url": str(request.url.with_query("").with_path(api.prefix))}],
    )

    spec = APISpec(
        options["info"].pop("title", f"{ api.app.cfg.name.title() } API"),
        options["info"].pop("version", "1.0.0"),
        options.pop("openapi_version", "3.0.0"),
        **options,
        plugins=[MarshmallowPlugin()],
    )
    tags = {}

    # Setup Authorization
    if api.authorize:
        _, _, schema = parse_docs(api.authorize)
        spec.options["security"] = []
        for key, value in schema.items():
            spec.components.security_scheme(key, value)
            spec.options["security"].append({key: []})

    # Setup Paths
    routes = api.router.routes()
    for route in routes:
        if route.path in SKIP_PATH:
            continue

        spec.path(route.path, **route_to_spec(route, spec, tags))

    return spec.to_dict()


def parse_docs(cb: Callable) -> Tuple[str, str, Dict]:
    """Parse docs from the given callback."""
    if yaml_utils is None:
        return "", "", {}

    docs = cb.__doc__ or ""
    schema = yaml_utils.load_yaml_from_docstring(docs)
    docs = docs.split("---")[0]
    docs = utils.dedent(utils.trim_docstring(docs))
    summary, _, description = docs.partition("\n\n")
    return summary, description.strip(), schema


def merge_dicts(source: Dict, merge: Dict) -> Dict:
    """Merge dicts."""
    return dict(
        source,
        **{
            key: (
                (
                    merge_dicts(source[key], merge[key])
                    if isinstance(source[key], dict) and isinstance(merge[key], dict)
                    else (
                        source[key] + merge[key]
                        if isinstance(source[key], list)
                        and isinstance(merge[key], list)
                        else merge[key]
                    )
                )
                if key in source
                else merge[key]
            )
            for key in merge
        },
    )


def route_to_spec(route: Route, spec: APISpec, tags: Dict) -> Dict:
    """Convert the given router to openapi operations."""
    results: Dict = {"parameters": [], "operations": {}}
    if isinstance(route, DynamicRoute):
        for param in route.params:
            results["parameters"].append({"in": "path", "name": param})

    target = cast(Callable, route.target)
    if isinstance(target, partial):
        target = target.func

    if hasattr(target, "openapi"):
        results["operations"] = target.openapi(route, spec, tags)
        return results

    summary, desc, schema = parse_docs(target)
    responses = return_type_to_response(target)
    for method in route_to_methods(route):
        results["operations"][method] = {
            "summary": summary,
            "description": desc,
            "responses": responses,
        }

    results["operations"] = merge_dicts(results["operations"], schema)
    return results


def route_to_methods(route: Route) -> List[str]:
    """Get sorted methods from the route."""
    methods = [m for m in HTTP_METHODS if m in (route.methods or [])]
    return [m.lower() for m in methods or DEFAULT_METHODS]


def return_type_to_response(fn: Callable) -> Dict:
    """Generate reponses specs based on the given function's return type."""
    responses: Dict[int, Dict] = {}
    return_type = fn.__annotations__.get("return")
    if return_type is None:
        return responses

    return_type = CAST_RESPONSE.get(return_type, return_type)
    if return_type is None:
        return responses

    if (
        inspect.isclass(return_type)
        and issubclass(return_type, Response)
        and return_type.content_type
    ):
        responses[return_type.status_code] = {
            "description": HTTPStatus(return_type.status_code).description,
            "content": {return_type.content_type: {}},
        }
    return responses


class OpenAPIMixin:
    """Render an endpoint to openapi specs."""

    meta: RESTOptions

    @classmethod
    def openapi(cls, route: Route, spec: APISpec, tags: Dict) -> Dict:  # noqa: C901
        """Get openapi specs for the endpoint."""
        meta = cls.meta
        if getattr(meta, meta.base_property, None) is None:
            return {}

        operations: Dict = {}
        summary, desc, schema = parse_docs(cls)
        if cls not in tags:
            tags[cls] = meta.name
            spec.tag({"name": meta.name, "description": summary})
            schema_cls = meta.Schema
            if schema_cls.__name__ not in spec.components.schemas:
                spec.components.schema(meta.Schema.__name__, schema=meta.Schema)

        schema_ref = {"$ref": f"#/components/schemas/{ meta.Schema.__name__ }"}
        for method in route_to_methods(route):
            operations[method] = {"tags": [tags[cls]]}
            is_resource_route = isinstance(route, DynamicRoute) and route.params.get(
                meta.name_id,
            )

            if method == "get" and not is_resource_route:
                operations[method]["parameters"] = []
                if meta.sorting:
                    operations[method]["parameters"].append(meta.sorting.openapi)

                if meta.filters:
                    operations[method]["parameters"].append(meta.filters.openapi)

                if meta.limit:
                    operations[method]["parameters"].append(
                        {
                            "name": LIMIT_PARAM,
                            "in": "query",
                            "schema": {
                                "type": "integer",
                                "minimum": 1,
                                "maximum": meta.limit,
                            },
                            "description": "The number of items to return",
                        },
                    )
                    operations[method]["parameters"].append(
                        {
                            "name": OFFSET_PARAM,
                            "in": "query",
                            "schema": {"type": "integer", "minimum": 0},
                            "description": "The offset of items to return",
                        },
                    )

            # Update from the method
            meth = getattr(cls, method, None)
            if (
                isinstance(route.target, partial)
                and "__meth__" in route.target.keywords
            ):
                meth = getattr(cls, route.target.keywords["__meth__"], None)

            elif method in {"post", "put"}:
                operations[method]["requestBody"] = {
                    "required": True,
                    "content": {"application/json": {"schema": schema_ref}},
                }

            if meth:
                (
                    operations[method]["summary"],
                    operations[method]["description"],
                    mschema,
                ) = openapi.parse_docs(
                    meth,
                )
                return_type = meth.__annotations__.get("return")
                if return_type in ("JSONType", TJSON):
                    responses = {
                        200: {
                            "description": "Request is successfull",
                            "content": {"application/json": {"schema": schema_ref}},
                        },
                    }
                else:
                    responses = return_type_to_response(meth)

                operations[method]["responses"] = responses
                operations[method] = merge_dicts(operations[method], mschema)

        return merge_dicts(operations, schema)



"""Support filters for Peewee ORM."""
from __future__ import annotations

import operator
from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, Union, cast

from peewee import Field, ModelSelect

from muffin_rest.filters import Filter, Filters

if TYPE_CHECKING:
    import marshmallow as ma


class PWFilter(Filter):
    """Support Peewee."""

    operators = Filter.operators
    operators["$in"] = operator.lshift
    operators["$none"] = operator.rshift
    operators["$like"] = operator.mod
    operators["$ilike"] = operator.pow
    operators["$contains"] = lambda f, v: f.contains(v)
    operators["$starts"] = lambda f, v: f.startswith(v)
    operators["$ends"] = lambda f, v: f.endswith(v)
    operators["$between"] = lambda f, v: f.between(*v)
    operators["$regexp"] = lambda f, v: f.regexp(v)

    list_ops = [*Filter.list_ops, "$between"]

    def __init__(
        self,
        name: str,
        *,
        field: Optional[Field] = None,
        schema_field: Optional[ma.fields.Field] = None,
        operator: Optional[str] = None,
        **_,
    ):
        """Support custom model fields."""
        self.name = name
        self.field = field
        self.schema_field = schema_field or self.schema_field_cls(
            attribute=field and field.name or name,
        )
        if operator:
            self.default_operator = operator

    async def filter(
        self, collection: ModelSelect, *ops: Tuple[Callable, Any], **kwargs,
    ) -> ModelSelect:
        """Apply the filters to Peewee QuerySet.."""
        if self.field and ops:
            return self.query(collection, self.field, *ops, **kwargs)
        return collection

    def query(self, qs: ModelSelect, column: Field, *ops: Tuple, **_) -> ModelSelect:
        """Filter a query."""
        if isinstance(column, Field):
            return qs.where(*[op(column, val) for op, val in ops])

        return qs


class PWFilters(Filters):
    """Bind Peewee filter class."""

    MUTATE_CLASS = PWFilter

    def convert(self, obj: Union[str, Field, PWFilter], **meta):
        """Convert params to filters."""
        from . import PWRESTHandler

        handler = cast(PWRESTHandler, self.handler)
        if isinstance(obj, PWFilter):
            if obj.field is None:
                obj.field = handler.meta.model._meta.fields.get(obj.name)
            return obj

        if isinstance(obj, Field):
            name = obj.name
            field = obj

        else:
            name = obj
            field = meta.pop("field", None) or name
            if isinstance(field, str):
                field = handler.meta.model._meta.fields.get(field)

        schema_field = meta.pop("schema_field", None)
        if schema_field is None and field:
            schema_field = handler.meta.Schema._declared_fields.get(field.name)
        return self.MUTATE_CLASS(name, field=field, schema_field=schema_field, **meta)



"""Implement sorting."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generator, List, Mapping, Sequence, Tuple, cast

from .types import TVCollection
from .utils import Mutate, Mutator

if TYPE_CHECKING:
    from muffin import Request
    from muffin.handler import Handler

SORT_PARAM = "sort"


class Sort(Mutate):
    """Sort a collection."""

    async def apply(self, collection, *, desc: bool = False, **_) -> Any:
        """Sort the collection."""
        return sorted(collection, key=lambda obj: getattr(obj, self.name), reverse=desc)


class Sorting(Mutator):

    """Build sorters for handlers."""

    MUTATE_CLASS = Sort
    mutations: Mapping[str, Sort]

    def __init__(self, handler: Handler, params: Sequence):
        """Initialize the sorting."""
        self.default: List[Sort] = []
        super(Sorting, self).__init__(handler, params)

    async def apply(
        self, request: Request, collection: TVCollection, **_,
    ) -> TVCollection:
        """Sort the given collection."""
        data = request.url.query.get(SORT_PARAM)
        if data:
            for name, desc in to_sort(data.split(",")):
                if name in self.mutations:
                    collection = await self.mutations[name].apply(collection, desc=desc)

        elif self.default:
            return self.sort_default(collection)

        return collection

    def convert(self, obj, **meta) -> Sort:
        """Prepare sorters."""
        sort = cast(Sort, super(Sorting, self).convert(obj, **meta))
        if sort.meta.get("default"):
            self.default.append(sort)
        return sort

    def sort_default(self, collection: TVCollection) -> TVCollection:
        """Sort by default."""
        return cast(TVCollection, sorted(collection))

    @property
    def openapi(self):
        """Prepare OpenAPI params."""
        sorting = list(self.mutations)
        return {
            "name": SORT_PARAM,
            "in": "query",
            "style": "form",
            "explode": False,
            "schema": {"type": "array", "items": {"type": "string", "enum": sorting}},
            "description": ",".join(sorting),
        }


def to_sort(
    sort_params: Sequence[str],
) -> Generator[Tuple[str, bool], None, None]:
    """Generate sort params."""
    for name in sort_params:
        n, desc = name.strip("-"), name.startswith("-")
        if n:
            yield n, desc



"""Base class for API REST Handlers."""

from __future__ import annotations

import abc
import inspect
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Generic,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from asgi_tools.response import parse_response
from asgi_tools.types import TJSON
from marshmallow import ValidationError
from muffin.handler import Handler, HandlerMeta

from muffin_rest import LIMIT_PARAM, OFFSET_PARAM, openapi
from muffin_rest.errors import APIError

from .errors import HandlerNotBindedError
from .options import RESTOptions
from .types import TVResource

if TYPE_CHECKING:
    import marshmallow as ma
    from muffin import Request

    from muffin_rest.api import API
    from muffin_rest.filters import Filter
    from muffin_rest.sorting import Sort


class RESTHandlerMeta(HandlerMeta):
    """Create class options."""

    def __new__(mcs, name, bases, params):
        """Prepare options for the handler."""
        kls = super().__new__(mcs, name, bases, params)
        kls.meta = kls.meta_class(kls)

        if getattr(kls.meta, kls.meta_class.base_property, None) is not None:
            kls.meta.filters = kls.meta.filters_cls(kls, kls.meta.filters)
            kls.meta.sorting = kls.meta.sorting_cls(kls, kls.meta.sorting)

        return kls


class RESTBase(Generic[TVResource], Handler, metaclass=RESTHandlerMeta):

    """Load/save resources."""

    auth: Any
    collection: Any
    resource: Any

    meta: RESTOptions
    meta_class: Type[RESTOptions] = RESTOptions
    _api: Optional[API] = None

    class Meta:
        """Tune the handler."""

        # Resource filters
        filters: Sequence[Union[str, Tuple[str, str], Filter]] = ()

        # Define allowed resource sorting params
        sorting: Sequence[Union[str, Tuple[str, Dict], Sort]] = ()

        # Serialize/Deserialize Schema class
        Schema: Optional[Type[ma.Schema]] = None

    @classmethod
    def __route__(cls, router, *paths, **params):
        """Bind the class to the given router."""
        methods = params.pop("methods") or cls.methods
        if paths:
            router.bind(cls, *paths, methods=methods, **params)

        else:
            router.bind(cls, f"/{ cls.meta.name }", methods=methods, **params)
            router.bind(
                cls,
                f"/{ cls.meta.name }/{{{ cls.meta.name_id }}}",
                methods=methods,
                **params,
            )

        for _, method in inspect.getmembers(cls, lambda m: hasattr(m, "__route__")):
            cpaths, cparams = method.__route__
            router.bind(cls, *cpaths, __meth__=method.__name__, **cparams)

        return cls

    async def __call__(
        self,
        request: Request,
        *,
        __meth__: Optional[str] = None,
        **options,
    ) -> Any:
        """Dispatch the given request by HTTP method."""
        method = getattr(self, __meth__ or request.method.lower())
        self.auth = await self.authorize(request)
        self.collection = await self.prepare_collection(request)
        resource = await self.prepare_resource(request)
        if resource or request.method != "GET":
            try:
                return await method(request, resource=resource)
            except ValidationError as exc:
                raise APIError.BAD_REQUEST("Invalid data", errors=exc.messages) from exc

        meta = self.meta

        # Filter collection
        if meta.filters:
            self.collection = await meta.filters.apply(
                request,
                self.collection,
                **options,
            )

        # Sort collection
        if meta.sorting:
            self.collection = await meta.sorting.apply(
                request,
                self.collection,
                **options,
            )

        # Paginate the collection
        headers = {}
        if meta.limit:
            limit, offset = self.paginate_prepare_params(request)
            if limit and offset >= 0:
                self.collection, total = await self.paginate(
                    request,
                    limit=limit,
                    offset=offset,
                )
                headers = self.paginate_prepare_headers(limit, offset, total)

        response = await method(request, resource=resource)

        if headers:
            response = parse_response(response)
            response.headers.update(headers)

        return response

    @property
    def api(self) -> API:
        """Check if the handler is binded to an API."""
        if self._api is None:
            raise HandlerNotBindedError
        return self._api

    async def authorize(self, request: Request) -> Any:
        """Default authorization method. Proxy auth to self.api."""
        auth = await self.api.authorize(request)
        if not auth:
            raise APIError.UNAUTHORIZED()
        return auth

    # Prepare data
    # ------------
    @abc.abstractmethod
    async def prepare_collection(self, request: Request) -> Any:
        """Prepare a collection of resources. Create queryset, db cursor and etc."""
        raise NotImplementedError

    async def prepare_resource(self, request: Request) -> Any:
        """Load a resource."""
        return request["path_params"].get(self.meta.name_id)

    # Paginate
    # --------
    def paginate_prepare_headers(self, limit, offset, total):
        """Prepare pagination headers."""
        return {"x-total": total, "x-limit": limit, "x-offset": offset}

    def paginate_prepare_params(self, request: Request) -> Tuple[int, int]:
        """Prepare pagination params."""
        meta = self.meta
        query = request.url.query
        limit = query.get(LIMIT_PARAM) or meta.limit
        try:
            return min(abs(int(limit)), meta.limit_max), int(query.get(OFFSET_PARAM, 0))
        except ValueError as exc:
            raise APIError.BAD_REQUEST("Pagination params are invalid") from exc

    @abc.abstractmethod
    async def paginate(
        self,
        request: Request,
        *,
        limit: int = 0,
        offset: int = 0,
    ) -> Tuple[Any, int]:
        """Paginate the results."""
        raise NotImplementedError

    # Manage data
    # -----------
    @abc.abstractmethod
    async def save(self, request: Request, resource: TVResource) -> TVResource:
        """Save the given resource."""
        return resource

    @abc.abstractmethod
    async def remove(self, request: Request, resource):
        """Remove the given resource."""
        raise NotImplementedError

    # Parse data
    # -----------
    async def get_schema(self, request: Request, **_) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        assert self.meta.Schema, "RESTHandler.meta.Schema is required."
        query = request.url.query
        return self.meta.Schema(
            only=query.get("schema_only"),
            exclude=query.get("schema_exclude", ()),
        )

    async def parse(self, request: Request):
        """Parse data from the given request."""
        try:
            return await request.data(raise_errors=True)
        except (ValueError, TypeError) as exc:
            raise APIError.BAD_REQUEST(str(exc)) from exc

    async def load(
        self,
        request: Request,
        resource: Optional[TVResource] = None,
    ) -> Any:
        """Load data from request and create/update a resource."""
        data = await self.parse(request)
        schema = await self.get_schema(request, resource=resource)
        if not schema:
            return data

        return schema.load(
            data,
            partial=resource is not None,
            many=isinstance(data, list),
        )

    async def dump(
        self,
        request: Request,
        data: Optional[Iterable] = None,
        resource: Optional[TVResource] = None,
        **dump_schema_opts,
    ) -> TJSON:
        """Serialize the given response."""
        schema = await self.get_schema(request, resource=resource)
        if schema:
            return schema.dump(
                resource if resource is not None else data,
                **dump_schema_opts,
            )

        return cast(TJSON, data)

    async def get(self, request: Request, *, resource: Optional[TVResource] = None):
        """Get a resource or a collection of resources.

        Specify a path param to load a resource.
        """
        if resource:
            return await self.dump(request, resource=resource)

        return await self.dump(request, data=self.collection, many=True)

    async def post(self, request: Request, *, resource: Optional[TVResource] = None):
        """Create a resource.

        The method accepts a single resource's data or a list of resources to create.
        """
        data = await self.load(request, resource)
        if isinstance(data, list):
            return await self.dump(
                request,
                data=[await self.save(request, res) for res in data],
                many=True,
            )

        res = await self.save(request, data)
        return await self.dump(request, resource=res)

    async def put(self, request: Request, *, resource: Optional[TVResource] = None):
        """Update a resource."""
        if resource is None:
            raise APIError.NOT_FOUND()

        return await self.post(request, resource=resource)

    async def delete(self, request: Request, resource: TVResource | None = None):
        """Delete a resource."""
        if resource is None:
            raise APIError.NOT_FOUND()

        return await self.remove(request, resource)


class RESTHandler(RESTBase, openapi.OpenAPIMixin):
    """Basic Handler Class."""


def to_sort(
    sort_params: Sequence[str],
) -> Generator[Tuple[str, bool], None, None]:
    """Generate sort params."""
    for name in sort_params:
        n, desc = name.strip("-"), name.startswith("-")
        if n:
            yield n, desc



"""REST Options."""

from typing import Dict, Type

import marshmallow as ma

from .filters import Filters
from .sorting import Sorting


class RESTOptions:
    """Handler Options."""

    name: str = ""
    name_id: str = "id"

    # limit: Paginate results (set to None for disable pagination)
    limit: int = 0
    limit_max: int = 0

    # Base class for filters
    filters: Filters
    filters_cls: Type[Filters] = Filters

    # Base class for sorting
    sorting: Sorting
    sorting_cls: Type[Sorting] = Sorting

    # Auto generation for schemas
    Schema: Type[ma.Schema]
    schema_base: Type[ma.Schema] = ma.Schema
    schema_fields: Dict = {}
    schema_meta: Dict = {}
    schema_unknown: str = ma.EXCLUDE

    base_property: str = "name"

    def __init__(self, cls):
        """Inherit meta options."""
        for base in reversed(cls.mro()):
            if hasattr(base, "Meta"):
                for k, v in base.Meta.__dict__.items():
                    if not k.startswith("_"):
                        setattr(self, k, v)

        if getattr(self, self.base_property, None) is not None:
            self.setup(cls)

    def setup(self, cls):
        """Setup the options."""
        self.name = self.name or cls.__name__.lower()
        if not self.Schema:
            name = self.name or "Unknown"
            self.Schema = type(
                name.title() + "Schema",
                (self.schema_base,),
                dict(self.schema_fields, Meta=self.setup_schema_meta(cls)),
            )

        if not self.limit_max:
            self.limit_max = self.limit

    def setup_schema_meta(self, _):
        """Generate meta for schemas."""
        return type(
            "Meta",
            (object,),
            dict({"unknown": self.schema_unknown}, **self.schema_meta),
        )

    def __repr__(self):
        """Represent self as a string."""
        return f"<Options {self.name}>"



"""Support sorting for Peewee ORM."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union, cast

from peewee import Field

from muffin_rest.sorting import Sort, Sorting

if TYPE_CHECKING:
    from .types import TVCollection


class PWSort(Sort):
    """Sorter for Peewee."""

    async def apply(
        self, collection: TVCollection, *, desc: bool = False, **_,
    ) -> TVCollection:
        """Sort the collection."""
        return collection.order_by_extend(self.field if not desc else self.field.desc())


class PWSorting(Sorting):
    """Sort Peewee ORM Queries."""

    MUTATE_CLASS = PWSort

    def convert(self, obj: Union[str, Field, PWSort], **meta):
        """Prepare sorters."""
        from . import PWRESTHandler

        if isinstance(obj, PWSort):
            return obj

        handler = cast(PWRESTHandler, self.handler)

        if isinstance(obj, Field):
            name, field = obj.name, obj

        else:
            name = obj
            field = meta.get("field", handler.meta.model._meta.fields.get(name))

        if field:
            sort = self.MUTATE_CLASS(name, field=field, **meta)
            if sort.meta.get("default"):
                self.default.append(sort)

            return sort

    def sort_default(self, collection: TVCollection) -> TVCollection:
        """Sort collection by default."""
        sorting = [
            sort.field.desc() if sort.meta["default"] == "desc" else sort.field
            for sort in self.default
        ]
        return collection.order_by(*sorting)



from typing import TYPE_CHECKING, Any, Awaitable, Callable, Type, TypeVar

if TYPE_CHECKING:
    from .handler import RESTBase

from muffin import Request

TVCollection = TypeVar("TVCollection", bound=Any)
TVResource = TypeVar("TVResource", bound=Any)
TAuth = Callable[[Request], Awaitable]
TVAuth = TypeVar("TVAuth", bound=TAuth)
TVHandler = TypeVar("TVHandler", bound=Type["RESTBase"])



from typing import Any, Dict, TypeVar

from .utils import MongoChain

TVCollection = TypeVar("TVCollection", bound=MongoChain)
TResource = Dict[str, Any]
TVResource = TypeVar("TVResource", bound=TResource)



"""Helpers to raise API errors as JSON responses."""
from __future__ import annotations

import json
from http import HTTPStatus
from typing import TYPE_CHECKING, Dict, Optional

from muffin import ResponseError

if TYPE_CHECKING:
    from asgi_tools.types import TJSON


class APIError(ResponseError):
    """JSON Response."""

    def __init__(
        self,
        content: Optional[TJSON] = None,
        *,
        status_code: int = HTTPStatus.BAD_REQUEST.value,
        **json_data,
    ):
        """Create JSON with errors information."""
        response = {"error": True, "message": HTTPStatus(status_code).description}

        if isinstance(content, Dict):
            response = content

        elif content is not None:
            response["message"] = str(content)

        if json_data:
            response.update(json_data)

        super(APIError, self).__init__(
            json.dumps(response),
            status_code=status_code,
            headers={"content-type": "application/json"},
        )


class InvalidEnpointError(TypeError):
    """Invalid endpoint."""

class HandlerNotBindedError(RuntimeError):
    """Handler not binded."""
